/* Proiectarea cu Microprocesoare - 2014
 *
 * Proiect - "Pedalometru" :D
 * ==== Ciclocomputer ===
 * 
 * Autor: Nedelcu Ana-Florentina, 334CC
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>

#include "lcd.h"

/* intreruperi:
 * senzor hall - pa7 - PCINT7
 * butoane 
	- inapoi - Pa0 - PCINT20
	- ok - Pa1 - PCINT1
	- jos - Pa2 - PCINT2
	- dreapta - Pa3 - PCINT3
	- stanga - Pa4 - PCINT4
	- sus - Pa5 - PCINT5
 */

#define PI 				3.141592;
#define line2_addr		0x40
#define ADC_CHANNEL 	7
#define no_m_field		510		// valoare determinata experimental
#define EPS				30

// transforma viteza din cm/s in km/h
#define 	transform_speed(v) (v * 0.036)

enum displayed_info {
	SPEED,
	DISTANCE,
	TIME,
	ROT
};

enum state {
	MENU,
	MEASURING
};


////////////////////////////////////////////////////////////////////////////////


int diam;			// in cm
int last_rot;
float speed = 0;		// in cm/s
int time;			// in secunde
float distance = 0;		// in km

enum displayed_info d_info;
enum state current_state;

int digit_is_displayed;
int current_digit;
int disp_diam[] = {0, 0, 0};
uint8_t digit_addr[] = {0x40, 0x42, 0x44}; 
int rot;


////////////////////////////////////////////////////////////////////////////////


int number_to_int(int* nr) {
	int i;

	i = nr[0] * 0 + nr[1] * 10 + nr[2];

	return i;
}

/* reafisez pe ecran */
void update_display() {

	char val[16];
	switch(d_info) {
		case SPEED:
			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("Viteza: ");
			
			int int_speed = transform_speed(speed);
			int zec_speed = ((float)transform_speed(speed) - int_speed) * 100;
			
			sprintf(val, "%d.%d km/h", int_speed, zec_speed * 100 );
			LCD_printAt(line2_addr, val);

			break;

		case DISTANCE:
			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("Distanta:");
			
			int int_dist = (int)distance;
			int zec_dist = (distance - int_dist) * 100;

			sprintf(val, "%d.%d km", int_dist, zec_dist);
			LCD_printAt(line2_addr, val);
			
			break;

		case TIME:
			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("Timp trecut: ");

			sprintf(val, "%d s", time);
			LCD_printAt(line2_addr, val);
			break;
			
		case ROT:
			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("Rotatii: ");

			sprintf(val, "%d rotatii", rot);
			LCD_printAt(line2_addr, val);
	}
} 


////////////////////////////////////////////////////////////////////////////////


/* rutina de tratare a intreruperilor cauzate de butoane - 
 * meniu interactiv in care se introduce diametrul 
 * si navigare intre informatiile afisate
 */
ISR(PCINT0_vect) {

	char val[1];

	if(current_state == MENU) {
		if((PINA & (1 << PA1)) == 0) {		// ok
			diam = number_to_int(disp_diam);
			current_state = MEASURING;
			update_display();
			time = 0;
		}

		else if((PINA & (1 << PA2)) == 0) {		// jos
			disp_diam[current_digit] = abs((disp_diam[current_digit] - 1 + 10) % 10);
			sprintf(val, "%d", disp_diam[current_digit]);			
			LCD_printAt(digit_addr[current_digit], val); 
			digit_is_displayed = 1;			
		} 

		else if((PINA & (1 << PA3)) == 0) {		// dreapta
			if(current_digit < 2)  {
				sprintf(val, "%d", disp_diam[current_digit]);			
				LCD_printAt(digit_addr[current_digit], val); 
				current_digit++;
				LCD_printAt(digit_addr[current_digit], " "); 
				digit_is_displayed = 0;
			}
		}
		
		else if((PINA & (1 << PA4)) == 0) {		// stanga
			if(current_digit > 0)  {
				sprintf(val, "%d", disp_diam[current_digit]);			
				LCD_printAt(digit_addr[current_digit], val); 
				current_digit--;
				LCD_printAt(digit_addr[current_digit], " "); 
				digit_is_displayed = 0;
			}
		}

		else if((PINA & (1 << PA5)) == 0) {		// sus
			disp_diam[current_digit] = (disp_diam[current_digit] + 1) % 10;
			sprintf(val, "%d", disp_diam[current_digit]);			
			LCD_printAt(digit_addr[current_digit], val); 
			digit_is_displayed = 1;	
		}
	} else {
		if((PINA & (1 << PA3)) == 0) {		// dreapta
			d_info = (d_info + 1) % 4;
			update_display();
		}

		else if((PINA & (1 << PA4)) == 0) {		// stanga
			d_info = (d_info - 1 + 4) % 4;
			update_display();
		}
	}
}

/* rutina de tratare a intreruperii generate de timerul 1 */
ISR(TIMER1_COMPA_vect)
{

	if(current_state == MENU) {

		if(digit_is_displayed == 0) {
			LCD_putCharAt(digit_addr[current_digit], (char)disp_diam[current_digit]);
			char val[1];
			sprintf(val, "%d", disp_diam[current_digit]);			
			LCD_printAt(digit_addr[current_digit], val); 
			digit_is_displayed = 1;
		} else {
			LCD_putCharAt(digit_addr[current_digit], ' ');
			digit_is_displayed = 0;
		}
	} else {
		float delta_d;
		int delta_rot;
		
		delta_rot = rot - last_rot;
		delta_d = delta_rot * 3.141592 * diam;
		distance += delta_d/100000;		// km
		speed = delta_d;

		last_rot = rot;
		
		time++;

		update_display();
	}
	
}

/* citire ADC */
int ADC_get(uint8_t channel)
{
	//setat canalul de pe care facem citirea
	ADMUX &= 0xE0;
	ADMUX |= ADC_CHANNEL;
	
	//pornim conversia
	ADCSRA |= (1 << ADSC);
	
	//asteptam sa se termine o conversie
	while(ADCSRA & (1 << ADSC));
	
	//returnam rezultatul conversiei	
	return ADC;
}


////////////////////////////////////////////////////////////////////////////////


/* initializare intrerupere pentru senzor */
void PCINT_sensor_init()
{
	PCICR |= (1 << PCIE0);
	PCMSK0 |= (1 << PCINT4);

}

// initializare intreruperi pt butoane
void PCINT_buttons_init() {
	PCICR |= (1 << PCIE0);
	PCMSK0 |= 0b00111111;
}


// initializez timer care genereaza intrerupere la 1s
void timer1_init()
{
	TCCR1B |= (1 << CS12)| (1 << CS10) | (1 << WGM12);

	TIMSK1 |= (1 << OCIE1A);

	OCR1A = 15625;
}

/* initializare intrari/iesiri */

void IO_init()
{
	/* LED-ul este pus ca iesire */
	DDRD |= (1 << PD7);

	/* butoanele si senzorul sunt puse ca intrari */
	DDRA &= ~( (1 << PA0) | (1 << PA1) | (1 << PA2) | 
				(1 << PA3) | (1 << PA4) | (1 << PA5) | (1 << PA7));
	PORTA |= (1 << PA0) | (1 << PA1) | (1 << PA2) | 
			(1 << PA3) | (1 << PA4) | (1 << PA5) | (1 << PA7);
}

/* initializare ADC */

void ADC_init(void)
{
	// fara intrerupere
	// prescaler la 128
	ADMUX = (1 << REFS0);
	ADCSRA = (1 << ADEN) | (7 << ADPS0);
}


////////////////////////////////////////////////////////////////////////////////


int main(void)
{
	char msg_l1[] = "Diametrul rotii:";
	int m_field, m_field_dif;

	 // meniu
	 current_state = MENU;
	 digit_is_displayed = 1;
	 current_digit = 0;
	 time = 0;
	 rot = 0;
	 last_rot= 0;
	 LCD_init();
	 IO_init();
	 timer1_init();

	 sei();

	/* - Afiseaza (introduceti diametrul)
		- ok => se afiseaza zerouri, primul clipeste
			(sageata stg-dr, se muta clipirea pe alta cifra, sageata sus-jos schimba cifra)
			ok => se citeste diametrul, se porneste vitezometrul
	*/

	LCD_print(msg_l1);
	LCD_printAt(line2_addr, "0 0 0");
	
	PCINT_buttons_init();

	/* Diametrul va fi in cm */

	while(1) {
		LCD_printAt(0x0f, " ");
		if(current_state == MEASURING) 	{
			break;
		}
		_delay_ms(10);

	}

	// initializari pentru vitezometru
	ADC_init();
	
	while(1) {
	
		m_field = ADC_get(ADC_CHANNEL);
		
		m_field_dif = m_field - no_m_field;
		
		if(m_field_dif < (-EPS) || m_field_dif > EPS) {
			
			// aprind ledul user de pe placa de baza
			PORTD |= 1 << PD7;
		
			while(m_field_dif < (-EPS) || m_field_dif > EPS) {	//asteptam sa se iasa
				//Citim valoarea data de senzor
				m_field = ADC_get(ADC_CHANNEL);
				m_field_dif = m_field - no_m_field;
			}
			rot++;
			update_display();

		}	
		
		PORTD &= ~(1 << PD7);

		_delay_ms(100);
	
	}

}

